﻿using System;
using System.IO;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using NiVE2.Drawing;
using ruche.nive2.effects;

using Drawing = System.Drawing;
using Imaging = System.Drawing.Imaging;
using Media = System.Windows.Media;
using Shapes = System.Windows.Shapes;

namespace ruche.nive2.effects.wpf
{
    /// <summary>
    /// ビットマップ関連の処理を定義する静的クラス。
    /// </summary>
    public static class BitmapUtil
    {
        #region 32bppサイズ計算

        /// <summary>
        /// 指定した幅と高さを持つ32bppイメージのデータサイズを算出する。
        /// </summary>
        /// <param name="width">幅。</param>
        /// <param name="height">高さ。</param>
        /// <returns>データサイズ。</returns>
        public static long Calc32bppSize(double width, double height)
        {
            return (4L * (long)(width + 0.5) * (long)(height + 0.5));
        }

        /// <summary>
        /// ソースイメージを32bppイメージに変換した時のデータサイズを算出する。
        /// </summary>
        /// <param name="src">ソースイメージ。</param>
        /// <returns>データサイズ。</returns>
        public static long Calc32bppSize(BitmapSource src)
        {
            return Calc32bppSize(src.PixelWidth, src.PixelHeight);
        }

        /// <summary>
        /// ソースイメージ32bppイメージに変換した時のデータサイズを算出する。
        /// </summary>
        /// <param name="src">ソースイメージ。</param>
        /// <returns>データサイズ。</returns>
        public static long Calc32bppSize(NBitmap src)
        {
            return Calc32bppSize(src.Width, src.Height);
        }

        /// <summary>
        /// ソースイメージを32bppイメージに変換した時のデータサイズを算出する。
        /// </summary>
        /// <param name="src">ソースイメージ。</param>
        /// <returns>データサイズ。</returns>
        public static long Calc32bppSize(Drawing.Image src)
        {
            return Calc32bppSize(src.Width, src.Height);
        }

        #endregion

        #region BGRA32形式 BitmapSource インスタンス生成

        /// <summary>
        /// RenderTargetBitmap インスタンスを生成する。
        /// </summary>
        /// <param name="width">幅。</param>
        /// <param name="height">高さ。</param>
        /// <returns>RenderTargetBitmap インスタンス。</returns>
        public static RenderTargetBitmap CreateRenderTargetBitmap(
            double width, double height)
        {
            return new RenderTargetBitmap(
                (int)(width + 0.5),
                (int)(height + 0.5),
                96,
                96,
                PixelFormats.Pbgra32);
        }

        /// <summary>
        /// RenderTargetBitmap インスタンスを生成する。
        /// </summary>
        /// <param name="width">幅。</param>
        /// <param name="height">高さ。</param>
        /// <param name="src">初期表示ビジュアル。</param>
        /// <returns>RenderTargetBitmap インスタンス。</returns>
        public static RenderTargetBitmap CreateRenderTargetBitmap(
            double width, double height, Visual src)
        {
            var dest = CreateRenderTargetBitmap(width, height);

            dest.Render(src);

            return dest;
        }

        /// <summary>
        /// RenderTargetBitmap インスタンスを生成する。
        /// </summary>
        /// <param name="width">幅。</param>
        /// <param name="height">高さ。</param>
        /// <param name="fillBrush">塗り潰すブラシ。</param>
        /// <returns>RenderTargetBitmap インスタンス。</returns>
        public static RenderTargetBitmap CreateRenderTargetBitmap(
            double width, double height, Media.Brush fillBrush)
        {
            var rect = new Shapes.Rectangle();
            rect.Width = width;
            rect.Height = height;
            rect.Fill = fillBrush;
            rect.Measure(new Size(width, height));
            rect.Arrange(new Rect(new Size(width, height)));

            return CreateRenderTargetBitmap(width, height, rect);
        }

        /// <summary>
        /// WriteableBitmap インスタンスを生成する。
        /// </summary>
        /// <param name="width">幅。</param>
        /// <param name="height">高さ。</param>
        /// <returns>WriteableBitmap インスタンス。</returns>
        public static WriteableBitmap CreateWriteableBitmap(
            double width, double height)
        {
            return new WriteableBitmap(
                (int)(width + 0.5),
                (int)(height + 0.5),
                96,
                96,
                PixelFormats.Bgra32,
                null);
        }

        /// <summary>
        /// WriteableBitmap インスタンスを生成する。
        /// </summary>
        /// <param name="src">ソースイメージ。</param>
        /// <returns>WriteableBitmap インスタンス。</returns>
        public static WriteableBitmap CreateWriteableBitmap(NBitmap src)
        {
            var dest = CreateWriteableBitmap(src.Width, src.Height);

            var rect = new Int32Rect(0, 0, src.Width, src.Height);
            dest.Lock();
            dest.WritePixels(rect, src.GetData(), src.Width * 4, 0);
            dest.AddDirtyRect(rect);
            dest.Unlock();

            return dest;
        }

        /// <summary>
        /// WriteableBitmap インスタンスを生成する。
        /// </summary>
        /// <param name="src">ソースイメージ。</param>
        /// <returns>WriteableBitmap インスタンス。</returns>
        public static WriteableBitmap CreateWriteableBitmap(Drawing.Bitmap src)
        {
            WriteableBitmap dest = null;

            using (var bmp = new NBitmap(src))
            {
                dest = CreateWriteableBitmap(bmp);
            }

            return dest;
        }

        #endregion

        #region BitmapSource からの変換

        /// <summary>
        /// BitmapSource インスタンスをBGRA32形式に変換する。
        /// </summary>
        /// <param name="src">BitmapSource インスタンス。</param>
        /// <returns>
        /// BGRA32形式の BitmapSource インスタンス。
        /// src が既にBGRA32形式である場合は src そのもの。
        /// </returns>
        public static BitmapSource ConvertToBgra32(BitmapSource src)
        {
            var dest = src;

            if (src.Format != PixelFormats.Bgra32)
            {
                var bmp = src.IsFrozen ?
                    src : (src.GetCurrentValueAsFrozen() as BitmapSource);
                dest = new FormatConvertedBitmap(
                    bmp,
                    PixelFormats.Bgra32,
                    null,
                    0.0);
                dest.Freeze();
            }

            return dest;
        }

        /// <summary>
        /// BitmapSource インスタンスを NBitmap インスタンスに変換する。
        /// </summary>
        /// <param name="src">BitmapSource インスタンス。</param>
        /// <returns>NBitmap インスタンス。</returns>
        public static NBitmap ConvertToNBitmap(BitmapSource src)
        {
            NBitmap dest = null;

            using (Drawing.Bitmap bmp = ConvertToBitmap(src))
            {
                dest = new NBitmap(bmp);
            }

            return dest;
        }

        /// <summary>
        /// BitmapSource インスタンスを Bitmap インスタンスに変換する。
        /// </summary>
        /// <param name="src">BitmapSource インスタンス。</param>
        /// <returns>Bitmap インスタンス。</returns>
        public static Drawing.Bitmap ConvertToBitmap(BitmapSource src)
        {
            var bmp = src.IsFrozen ?
                src : (src.GetCurrentValueAsFrozen() as BitmapSource);

            var encoder = new PngBitmapEncoder();
            encoder.Frames.Add(BitmapFrame.Create(bmp));

            Drawing.Bitmap dest = null;
            using (var s = new MemoryStream())
            {
                encoder.Save(s);
                s.Seek(0, SeekOrigin.Begin);
                dest = new Drawing.Bitmap(s);
            }

            return dest;
        }

        #endregion
    }
}
