﻿using System;
using System.Windows.Media;
using NiVE2.Plugin.Interface;
using NiVE2.Plugin.Utils;

using Drawing2D = System.Drawing.Drawing2D;

namespace ruche.nive2.effects.wpf
{
    /// <summary>
    /// WPFのブラシを保持するNiVE2プロパティクラス。
    /// </summary>
    [Serializable]
    public class BrushProperty : PropertyBase
    {
        /// <summary>
        /// ブラシ。
        /// </summary>
        private XamlSerializable<Brush> _brush = new XamlSerializable<Brush>();

        /// <summary>
        /// コンストラクタ。
        /// </summary>
        /// <param name="name">プロパティ名。</param>
        public BrushProperty(string name) : this(name, null)
        {
        }

        /// <summary>
        /// コンストラクタ。
        /// </summary>
        /// <param name="name">プロパティ名。</param>
        /// <param name="brush">ブラシ。</param>
        public BrushProperty(string name, Brush brush) : base(name)
        {
            Brush = brush;
        }

        /// <summary>
        /// コピーコンストラクタ。
        /// </summary>
        /// <param name="src">コピー元。</param>
        public BrushProperty(BrushProperty src)
            :
            base(src.PropertyName)
        {
            Brush = src.Brush;
        }

        /// <summary>
        /// 既定のブラシを取得する。
        /// </summary>
        public static Brush DefaultBrush
        {
            get { return Brushes.Black; }
        }

        /// <summary>
        /// ブラシを取得または設定する。
        /// null を渡すと DefaultBrush が設定される。
        /// </summary>
        /// <remarks>
        /// 値の設定時にはクローンが作成され、即座にフリーズされる。
        /// </remarks>
        public Brush Brush
        {
            get { return _brush.Value; }
            set
            {
                SafeInvoker.Call(value, new Action(() =>
                {
                    _brush.Value = (value ?? DefaultBrush).Clone();
                    _brush.Value.Freeze();
                }));
            }
        }

        public override string ToString()
        {
            return Brush.ToString();
        }

        #region PropertyBase メンバ

        public override bool CanUseScript
        {
            get { return false; }
        }

        public override object Value
        {
            get { return Brush; }
            set { Brush = (Brush)value; }
        }

        public override PropertyInterpolationType SupportInterpolationType
        {
            get
            {
                return (
                    PropertyInterpolationType.Fixed |
                    PropertyInterpolationType.Liner);
            }
        }

        public override PropertyBase Interpolation(
            KeyFrame[] keyFrame,
            double time)
        {
            KeyFrame prevKey = Util.GetPrevKeyFrame(keyFrame, time);
            KeyFrame nextKey = Util.GetNextKeyFrame(keyFrame, time);

            PropertyBase prop = null;
            if (
                prevKey != null &&
                nextKey != null &&
                prevKey.Type == PropertyInterpolationType.Liner)
            {
                Brush brush = Interpolater.Linear(
                    ((BrushProperty)prevKey.Property).Brush,
                    ((BrushProperty)nextKey.Property).Brush,
                    (time - prevKey.Time) / (nextKey.Time - prevKey.Time));
                prop = new BrushProperty(this.PropertyName, brush);
            }
            else
            {
                prop = (prevKey ?? nextKey).Property.Copy();
            }

            return prop;
        }

        public override PropertyBase Copy()
        {
            return new BrushProperty(this);
        }

        public override PropertyBase PasteProperty(PropertyBase property)
        {
            return property;
        }

        public override PropertyBase[] PasteProperty(PropertyBase[] property)
        {
            return property;
        }

        public override bool Equals(PropertyBase obj)
        {
            if (obj is BrushProperty)
            {
                return (Brush == null) ?
                    (((BrushProperty)obj).Brush == null) :
                    Brush.Equals(((BrushProperty)obj).Brush);
            }
            return false;
        }

        #endregion
    }
}
