﻿using System;
using System.ComponentModel;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Forms;
using NiVE2.Plugin.Interface;
using NiVE2.Plugin.Utils;
using NiVE2.Utils;

using Drawing = System.Drawing;

namespace ruche.nive2.effects.wpf
{
    /// <summary>
    /// ブラシプロパティ値を設定するNiVE2プロパティエディットコントロールクラス。
    /// </summary>
    public class BrushPropertyEditControl : PropertyEditControlBase
    {
        private PictureBox pictureEdit;

        /// <summary>
        /// ブラシ種別配列。
        /// </summary>
        private static readonly EditBrushTypes[] BrushTypes =
            {
                EditBrushTypes.SolidColor,
                EditBrushTypes.LinearGradient,
                EditBrushTypes.RadialGradient,
            };

        /// <summary>
        /// ブラシ種別配列に対応するブラシクラス型配列。
        /// </summary>
        private static readonly Type[] BrushClassTypes =
            {
                typeof(SolidColorBrush),
                typeof(LinearGradientBrush),
                typeof(RadialGradientBrush),
            };

        /// <summary>
        /// ブラシ種別配列に対応するメニュー表示名配列。
        /// </summary>
        private static readonly string[] BrushMenuTexts =
            {
                "単一色",
                "線形グラデーション",
                "放射状グラデーション",
            };

        /// <summary>
        /// メニュー。
        /// </summary>
        private ContextMenuStrip _menu = null;

        /// <summary>
        /// 単一色ブラシ用カラーダイアログのカスタムカラー配列。
        /// </summary>
        private int[] _solidCustomColors =
            {
                0xFFFFFF, 0xFFFF00, 0xFF00FF, 0x00FFFF,
                0x008080, 0x800080, 0x808000, 0x808080,
                0x000000, 0x0000FF, 0x00FF00, 0xFF0000,
                0x800000, 0x008000, 0x000080, 0xC0C0C0,
            };

        /// <summary>
        /// 最後に選択したブラシ配列。
        /// </summary>
        private Brush[] _lastBrushes = new Brush[BrushClassTypes.Length];

        /// <summary>
        /// コンストラクタ。
        /// </summary>
        /// <param name="name">対応するプロパティ名。</param>
        public BrushPropertyEditControl(string name)
            :
            this(name, EditBrushTypes.All)
        {   
        }

        /// <summary>
        /// コンストラクタ。
        /// </summary>
        /// <param name="name">対応するプロパティ名。</param>
        /// <param name="types">編集可能なブラシ種別の組み合わせ。</param>
        public BrushPropertyEditControl(string name, EditBrushTypes types)
        {
            PropertyName = name;
            LabelName = name;

            // コントロール設定
            InitializeComponent();

            // ピクチャビュー背景画像変更
            InitializePictureBackground();

            // メニュー初期化
            InitializeMenu(types);
        }

        /// <summary>
        /// ピクチャビューの背景画像を初期化する。
        /// </summary>
        private void InitializePictureBackground()
        {
            // 画像作成
            var bmp = new Drawing.Bitmap(8, 8);
            using (var g = Drawing.Graphics.FromImage(bmp))
            {
                g.FillRectangles(
                    Drawing.Brushes.Silver,
                    new Drawing.Rectangle[]
                    {
                        new Drawing.Rectangle(0, 0, 4, 4),
                        new Drawing.Rectangle(4, 4, 4, 4),
                    });
                g.FillRectangles(
                    Drawing.Brushes.White,
                    new Drawing.Rectangle[]
                    {
                        new Drawing.Rectangle(0, 4, 4, 4),
                        new Drawing.Rectangle(4, 0, 4, 4),
                    });
            }

            // 設定
            pictureEdit.BackgroundImageLayout = ImageLayout.Tile;
            var temp = pictureEdit.BackgroundImage;
            pictureEdit.BackgroundImage = bmp;
            if (temp != null)
            {
                temp.Dispose();
            }
        }

        /// <summary>
        /// メニューを初期化する。
        /// </summary>
        /// <param name="types">編集可能なブラシ種別の組み合わせ。</param>
        private void InitializeMenu(EditBrushTypes types)
        {
            _menu = new ContextMenuStrip();

            // アイテム追加(Tag にクラス型設定)
            for (int i = 0; i < BrushTypes.Length; ++i)
            {
                if ((types & BrushTypes[i]) != 0)
                {
                    var item = new ToolStripMenuItem(BrushMenuTexts[i]);
                    item.Tag = BrushClassTypes[i];
                    _menu.Items.Add(item);
                }
            }

            // イベントハンドラ設定
            _menu.ItemClicked += menu_ItemClicked;
        }

        /// <summary>
        /// ブラシ編集ダイアログを表示する。
        /// </summary>
        /// <param name="type">ブラシクラス型。</param>
        /// <returns>作成されたブラシ。キャンセルされた場合は null 。</returns>
        private Brush ShowBrushDialog(Type type)
        {
            int index = Array.FindIndex(BrushClassTypes, (t) => (t == type));
            Brush lastBrush = (index < 0) ? null : _lastBrushes[index];

            Brush result = null;
            if (type == typeof(SolidColorBrush))
            {
                using (var dialog = new SolidColorBrushDialog())
                {
                    dialog.Brush = lastBrush as SolidColorBrush;
                    dialog.UserColor = dialog.BrushColor;
                    if (dialog.ShowDialog(this) == DialogResult.OK)
                    {
                        result = dialog.Brush;
                    }
                }
            }
            else if (type == typeof(LinearGradientBrush))
            {
                using (var dialog = new LinearGradientBrushDialog())
                {
                    dialog.Brush = lastBrush as LinearGradientBrush;
                    if (dialog.ShowDialog(this) == DialogResult.OK)
                    {
                        result = dialog.Brush;
                    }
                }
            }
            else if (type == typeof(RadialGradientBrush))
            {
                using (var dialog = new RadialGradientBrushDialog())
                {
                    dialog.Brush = lastBrush as RadialGradientBrush;
                    if (dialog.ShowDialog(this) == DialogResult.OK)
                    {
                        result = dialog.Brush;
                    }
                }
            }

            return result;
        }

        #region PropertyEditControlBase メンバ

        public override int SpliterDistance
        {
            get
            {
                return base.SpliterDistance;
            }
            set
            {
                base.SpliterDistance = value;

                // コントロールの位置合わせを行う
                pictureEdit.Left = value + Util.PropertyEditControlLeftDistance;
            }
        }

        public override Type UseProperetyType()
        {
            return typeof(BrushProperty);
        }

        public override void ChangeProperty(PropertyBase property)
        {
            var prop = property as BrushProperty;
            if (prop != null)
            {
                // ブラシからビットマップ作成
                var bmp = BitmapUtil.CreateRenderTargetBitmap(
                    pictureEdit.ClientSize.Width,
                    pictureEdit.ClientSize.Height,
                    prop.Brush);
                if (bmp.CanFreeze)
                {
                    bmp.Freeze();
                }
                var image = BitmapUtil.ConvertToBitmap(bmp);

                // ボタン画像に設定
                var temp = pictureEdit.Image;
                pictureEdit.Image = image;
                if (temp != null)
                {
                    temp.Dispose();
                }

                // 型検索＆選択ブラシ更新
                Type type = null;
                for (int i = 0; i < BrushClassTypes.Length; ++i)
                {
                    if (BrushClassTypes[i].IsInstanceOfType(prop.Brush))
                    {
                        type = BrushClassTypes[i];
                        _lastBrushes[i] = prop.Brush;
                        break;
                    }
                }

                // デフォルトメニュー項目変更
                foreach (ToolStripMenuItem item in _menu.Items)
                {
                    item.Checked = (type == (item.Tag as Type));
                }
            }

            base.ChangeProperty(property);
        }

        protected override void Dispose(bool disposing)
        {
            if (pictureEdit != null)
            {
                if (pictureEdit.Image != null)
                {
                    pictureEdit.Image.Dispose();
                    pictureEdit.Image = null;
                }
                if (pictureEdit.BackgroundImage != null)
                {
                    pictureEdit.BackgroundImage.Dispose();
                    pictureEdit.BackgroundImage = null;
                }
            }
            if (_menu != null)
            {
                _menu.Dispose();
                _menu = null;
            }

            base.Dispose(disposing);
        }

        #endregion

        private void InitializeComponent()
        {
            this.pictureEdit = new System.Windows.Forms.PictureBox();
            ((System.ComponentModel.ISupportInitialize)(this.pictureEdit)).BeginInit();
            this.SuspendLayout();
            // 
            // pictureEdit
            // 
            this.pictureEdit.BorderStyle = System.Windows.Forms.BorderStyle.FixedSingle;
            this.pictureEdit.Location = new System.Drawing.Point(114, 0);
            this.pictureEdit.Name = "pictureEdit";
            this.pictureEdit.Size = new System.Drawing.Size(38, 19);
            this.pictureEdit.TabIndex = 3;
            this.pictureEdit.TabStop = false;
            this.pictureEdit.Click += new System.EventHandler(this.pictureEdit_Click);
            // 
            // BrushPropertyEditControl
            // 
            this.AutoScaleDimensions = new System.Drawing.SizeF(6F, 12F);
            this.Controls.Add(this.pictureEdit);
            this.Name = "BrushPropertyEditControl";
            this.Controls.SetChildIndex(this.pictureEdit, 0);
            ((System.ComponentModel.ISupportInitialize)(this.pictureEdit)).EndInit();
            this.ResumeLayout(false);

        }

        private void pictureEdit_Click(object sender, EventArgs e)
        {
            if (_menu.Items.Count == 1)
            {
                // 編集可能ブラシが1種類なら即編集
                Brush brush = ShowBrushDialog(_menu.Items[0].Tag as Type);
                if (brush != null)
                {
                    BrushProperty prop = new BrushProperty(PropertyName, brush);
                    SetProperty(prop);
                }
            }
            else if (_menu.Items.Count > 1)
            {
                // 編集可能ブラシが2種類以上あるならメニュー表示
                _menu.Show(pictureEdit, pictureEdit.PointToClient(Cursor.Position));
            }
        }

        private void menu_ItemClicked(object sender, ToolStripItemClickedEventArgs e)
        {
            // 選択されたブラシ種別を編集
            Brush brush = ShowBrushDialog(e.ClickedItem.Tag as Type);
            if (brush != null)
            {
                BrushProperty prop = new BrushProperty(PropertyName, brush);
                SetProperty(prop);
            }
        }
    }
}
