﻿using System;
using System.Windows.Threading;
using System.Threading;

namespace ruche.nive2.effects.wpf
{
    /// <summary>
    /// スレッドが異なる場合のみ Dispatcher の Invoke 呼び出しを行う
    /// 静的メソッドを提供するクラス。
    /// </summary>
    public static class SafeInvoker
    {
        /// <summary>
        /// ベーススレッドの Dispatcher 。
        /// </summary>
        private static Dispatcher _baseDispatcher = null;

        /// <summary>
        /// ベーススレッドの Dispatcher を取得する。
        /// </summary>
        public static Dispatcher BaseDispatcher
        {
            get { return _baseDispatcher; }
        }

        /// <summary>
        /// ベーススレッドからの通知メソッド。
        /// 初回呼び出し時に BaseDispatcher を設定する。
        /// </summary>
        public static void NotifyBaseThread()
        {
            if (_baseDispatcher == null)
            {
                _baseDispatcher = Dispatcher.CurrentDispatcher;
            }
        }

        /// <summary>
        /// BaseDispatcher とスレッドが異なる場合のみ Invoke 呼び出しを行う。
        /// </summary>
        /// <param name="method">デリゲート。</param>
        /// <param name="args">引数リストもしくは null 。</param>
        /// <returns>デリゲートの戻り値もしくは null 。</returns>
        /// <remarks>
        /// BaseDispatcher と同一のスレッドから呼び出した場合は
        /// 通常のデリゲート呼び出しを行う。
        /// BaseDispatcher と異なるスレッドから呼び出した場合は
        /// Dispatcher の Invoke 呼び出しを行う。
        /// BaseDispatcher が未設定である場合は例外となる。
        /// </remarks>
        public static object Call(Delegate method, params object[] args)
        {
            return Call(BaseDispatcher, method, args);
        }

        /// <summary>
        /// スレッドが異なる場合のみ Dispatcher の Invoke 呼び出しを行う。
        /// </summary>
        /// <param name="dispatcher">Dispatcher 。</param>
        /// <param name="method">デリゲート。</param>
        /// <param name="args">引数リストもしくは null 。</param>
        /// <returns>デリゲートの戻り値もしくは null 。</returns>
        /// <remarks>
        /// dispatcher と同一のスレッドから呼び出した場合は
        /// 通常のデリゲート呼び出しを行う。
        /// dispatcher と異なるスレッドから呼び出した場合は
        /// Dispatcher の Invoke 呼び出しを行う。
        /// dispatcher が null である場合は BaseDispatcher を用いて判断する。
        /// BaseDispatcher が未設定である場合は例外となる。
        /// </remarks>
        public static object Call(
            Dispatcher dispatcher,
            Delegate method,
            params object[] args)
        {
            object result = null;

            if (dispatcher == null)
            {
                dispatcher = BaseDispatcher;
            }

            if (dispatcher.Thread == Thread.CurrentThread)
            {
                result = method.DynamicInvoke(args);
            }
            else
            {
                result = dispatcher.Invoke(method, args);
            }

            return result;
        }

        /// <summary>
        /// スレッドが異なる場合のみ Dispatcher の Invoke 呼び出しを行う。
        /// </summary>
        /// <param name="obj">DispatcherObject もしくは null 。</param>
        /// <param name="method">デリゲート。</param>
        /// <param name="args">引数リストもしくは null 。</param>
        /// <returns>デリゲートの戻り値もしくは null 。</returns>
        /// <remarks>
        /// obj と同一のスレッドから呼び出した場合は
        /// 通常のデリゲート呼び出しを行う。
        /// obj と異なるスレッドから呼び出した場合は
        /// Dispatcher の Invoke 呼び出しを行う。
        /// obj が null である場合は BaseDispatcher を用いて判断する。
        /// BaseDispatcher が未設定である場合は例外となる。
        /// </remarks>
        public static object Call(
            DispatcherObject obj,
            Delegate method,
            params object[] args)
        {
            return Call(
                (obj == null) ? BaseDispatcher : obj.Dispatcher,
                method,
                args);
        }
    }
}
