﻿using System;
using System.Windows.Media;
using System.Windows.Forms;

using Drawing = System.Drawing;
using Imaging = System.Drawing.Imaging;

namespace ruche.nive2.effects.wpf
{
    /// <summary>
    /// 
    /// </summary>
    public partial class SolidColorBrushDialog : Form
    {
        /// <summary>
        /// テンプレート色配列要素数。
        /// </summary>
        public const int TemplateColorCount = 16;

        /// <summary>
        /// 既定のテンプレート色配列。
        /// </summary>
        private static readonly Color[] _defaultTemplateColors =
            {
                Color.FromRgb(0xFF, 0xFF, 0xFF),
                Color.FromRgb(0xFF, 0xFF, 0x00),
                Color.FromRgb(0xFF, 0x00, 0xFF),
                Color.FromRgb(0x00, 0xFF, 0xFF),
                Color.FromRgb(0xC0, 0xC0, 0xC0),
                Color.FromRgb(0x80, 0x00, 0x80),
                Color.FromRgb(0x80, 0x80, 0x00),
                Color.FromRgb(0x00, 0x80, 0x80),
                Color.FromRgb(0x80, 0x80, 0x80),
                Color.FromRgb(0xFF, 0x00, 0x00),
                Color.FromRgb(0x00, 0xFF, 0x00),
                Color.FromRgb(0x00, 0x00, 0xFF),
                Color.FromRgb(0x00, 0x00, 0x00),
                Color.FromRgb(0x80, 0x00, 0x00),
                Color.FromRgb(0x00, 0x80, 0x00),
                Color.FromRgb(0x00, 0x00, 0x80),
            };

        /// <summary>
        /// ブラシ。
        /// </summary>
        private SolidColorBrush _brush = DefaultBrush;

        /// <summary>
        /// テンプレート色配列。
        /// </summary>
        private Color[] _templateColors = DefaultTemplateColors;

        /// <summary>
        /// テンプレート色ボタン配列。
        /// </summary>
        private Button[] _buttonDefs = null;

        /// <summary>
        /// 現在の色。
        /// </summary>
        private Color _color = Colors.Black;

        /// <summary>
        /// カラーダイアログのカスタムカラー配列。
        /// </summary>
        private int[] _customColors = new int[]
            {
                0xFFFFFF, 0xFFFF00, 0xFF00FF, 0x00FFFF,
                0x008080, 0x800080, 0x808000, 0x808080,
                0x000000, 0x0000FF, 0x00FF00, 0xFF0000,
                0x800000, 0x008000, 0x000080, 0xC0C0C0,
            };

        /// <summary>
        /// コンストラクタ。
        /// </summary>
        public SolidColorBrushDialog()
        {
            // コントロール初期化
            InitializeComponent();

            // ボタン配列作成
            _buttonDefs = new Button[]
                {
                    buttonDef0,
                    buttonDef1,
                    buttonDef2,
                    buttonDef3,
                    buttonDef4,
                    buttonDef5,
                    buttonDef6,
                    buttonDef7,
                    buttonDef8,
                    buttonDef9,
                    buttonDef10,
                    buttonDef11,
                    buttonDef12,
                    buttonDef13,
                    buttonDef14,
                    buttonDef15,
                };

            // ユーザ定義色初期化
            UserColor = DefaultBrush.Color;
        }

        /// <summary>
        /// 既定のブラシを取得する。
        /// </summary>
        public static SolidColorBrush DefaultBrush
        {
            get { return Brushes.Black; }
        }

        /// <summary>
        /// 既定のテンプレート色配列を取得する。
        /// </summary>
        public static Color[] DefaultTemplateColors
        {
            get { return _defaultTemplateColors.Clone() as Color[]; }
        }

        /// <summary>
        /// ブラシを取得または設定する。
        /// null を渡すと DefaultBrush が設定される。
        /// </summary>
        public SolidColorBrush Brush
        {
            get { return _brush; }
            set { _brush = value ?? DefaultBrush; }
        }

        /// <summary>
        /// ブラシ色を取得または設定する。
        /// </summary>
        public Color BrushColor
        {
            get { return Brush.Color; }
            set { Brush.Color = value; }
        }

        /// <summary>
        /// ブラシ色を System.Drawing.Color 型で取得または設定する。
        /// </summary>
        public Drawing.Color BrushDrawingColor
        {
            get { return (Drawing.Color)(Color32)BrushColor; }
            set { BrushColor = (Color32)value; }
        }

        /// <summary>
        /// テンプレート色配列を取得または設定する。
        /// null を渡すと DefaultTemplateColors が設定される。
        /// 実際に使われる色は GetTemplateColor メソッドで取得できる。
        /// </summary>
        public Color[] TemplateColors
        {
            get { return _templateColors; }
            set { _templateColors = value ?? DefaultTemplateColors; }
        }

        /// <summary>
        /// ユーザ定義色を取得または設定する。
        /// 設定時にアルファ成分値は取り除かれる。
        /// </summary>
        public Color UserColor
        {
            get { return new Color32(255, buttonDefUser.BackColor); }
            set
            {
                buttonDefUser.BackColor =
                    (Drawing.Color)new Color32(255, value);
            }
        }

        /// <summary>
        /// 実際のテンプレート色を取得する。
        /// </summary>
        /// <param name="index">インデックス。</param>
        /// <returns>テンプレート色。</returns>
        public Color GetTemplateColor(int index)
        {
            if (index < 0 || index >= TemplateColorCount)
            {
                throw new ArgumentOutOfRangeException(
                    "index",
                    index,
                    "index には 0 から " + (TemplateColorCount - 1) +
                    "までの値のみ指定できます。");
            }

            Color c = (index >= TemplateColors.Length) ?
                DefaultTemplateColors[index] :
                TemplateColors[index];
            c.A = 255;

            return c;
        }

        /// <summary>
        /// 色選択ダイアログを表示する。
        /// </summary>
        /// <param name="color">初期選択色。</param>
        /// <returns>選択された色。キャンセルされた場合は null 。</returns>
        private Drawing.Color? ShowColorDialog(Drawing.Color color)
        {
            Drawing.Color? result = null;

            using (var dialog = new ColorDialog())
            {
                dialog.Color = color;
                dialog.AnyColor = true;
                dialog.FullOpen = true;
                dialog.CustomColors = _customColors;
                if (dialog.ShowDialog(this) == DialogResult.OK)
                {
                    result = dialog.Color;
                    _customColors = dialog.CustomColors;
                }
            }

            return result;
        }

        /// <summary>
        /// ビューボックスの背景画像を初期化する。
        /// </summary>
        private void InitializeViewBackground()
        {
            // 画像作成
            var bmp = new Drawing.Bitmap(16, 16);
            using (var g = Drawing.Graphics.FromImage(bmp))
            {
                g.FillRectangles(
                    Drawing.Brushes.Silver,
                    new Drawing.Rectangle[]
                    {
                        new Drawing.Rectangle(0, 0, 8, 8),
                        new Drawing.Rectangle(8, 8, 8, 8),
                    });
                g.FillRectangles(
                    Drawing.Brushes.White,
                    new Drawing.Rectangle[]
                    {
                        new Drawing.Rectangle(0, 8, 8, 8),
                        new Drawing.Rectangle(8, 0, 8, 8),
                    });
            }

            // 設定
            pictureView.BackgroundImageLayout = ImageLayout.Tile;
            var temp = pictureView.BackgroundImage;
            pictureView.BackgroundImage = bmp;
            if (temp != null)
            {
                temp.Dispose();
            }
        }

        /// <summary>
        /// 現在の選択色を更新し、ビューボックス画像を変更する。
        /// </summary>
        /// <param name="color">更新色。アルファ成分値は無視される。</param>
        private void UpdateCurrentColor(Color color)
        {
            // 色設定
            _color = color;
            _color.A = (byte)updownAlpha.Value;

            // Bitmap 作成
            var bmp = new Drawing.Bitmap(
                pictureView.ClientSize.Width,
                pictureView.ClientSize.Height,
                Imaging.PixelFormat.Format32bppArgb);
            using (var g = Drawing.Graphics.FromImage(bmp))
            using (
                var db = new Drawing.SolidBrush((Drawing.Color)(Color32)_color))
            {
                g.FillRectangle(db, pictureView.ClientRectangle);
            }

            // ビュー画像変更
            var temp = pictureView.Image;
            pictureView.Image = bmp;
            if (temp != null)
            {
                temp.Dispose();
            }
        }

        private void SolidColorBrushDialog_Load(object sender, EventArgs e)
        {
            InitializeViewBackground();

            // テンプレート色設定
            for (int i = 0; i < TemplateColorCount; ++i)
            {
                _buttonDefs[i].BackColor =
                    (Drawing.Color)(Color32)GetTemplateColor(i);
            }

            // 不透明度設定
            updownAlpha.Value = BrushColor.A;

            // 色更新
            UpdateCurrentColor(BrushColor);
        }

        private void SolidColorBrushDialog_FormClosed(
            object sender, FormClosedEventArgs e)
        {
            if (DialogResult == DialogResult.OK)
            {
                // ブラシを反映
                Brush = new SolidColorBrush(_color);
            }

            // イメージを破棄
            if (pictureView.Image != null)
            {
                pictureView.Image.Dispose();
                pictureView.Image = null;
            }
            if (pictureView.BackgroundImage != null)
            {
                pictureView.BackgroundImage.Dispose();
                pictureView.BackgroundImage = null;
            }
        }

        private void buttonDef_Click(object sender, EventArgs e)
        {
            var button = sender as Button;
            if (button != null)
            {
                UpdateCurrentColor((Color32)button.BackColor);
            }
        }

        private void buttonDef_BackColorChanged(object sender, EventArgs e)
        {
            var button = sender as Button;
            if (button != null)
            {
                // ボタン文字列変更
                var c = button.BackColor;
                button.Text =
                    "#" + string.Format("{0:X2}{1:X2}{2:X2}", c.R, c.G, c.B);

                // 前景色変更
                button.ForeColor = (ImageUtil.CalcBrightness(c) < 0.5) ?
                    Drawing.Color.White : Drawing.Color.Black;
            }
        }

        private void buttonColor_Click(object sender, EventArgs e)
        {
            var color = ShowColorDialog((Drawing.Color)new Color32(255, _color));
            if (color != null)
            {
                Color c = (Color32)(Drawing.Color)color;
                UserColor = c;
                UpdateCurrentColor(c);
            }
        }

        private void updownAlpha_ValueChanged(object sender, EventArgs e)
        {
            UpdateCurrentColor(_color);
        }
    }
}
