﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.ComponentModel;
using System.Reflection;
using NiVE2.Plugin.Interface;
using NiVE2.Plugin.Utils;

namespace ruche.nive2.effects
{
    /// <summary>
    /// 便利処理を定義する静的クラス。
    /// </summary>
    public static class Util
    {
        /// <summary>
        /// プロパティエディットコントロールのスプリッタ右端からの基準位置。
        /// </summary>
        public const int PropertyEditControlLeftDistance = 6;

        /// <summary>
        /// プロパティエディットコントロールの名前を分割文字列で分割し、
        /// 一番最後の文字列をラベル文字列に設定する。
        /// </summary>
        /// <param name="control">プロパティエディットコントロール。</param>
        /// <param name="splitter">分割文字列。</param>
        public static void SetPropertyEditControlLabelByName(
            PropertyEditControlBase control, string splitter)
        {
            int index = control.PropertyName.LastIndexOf(splitter);
            control.LabelName = (index < 0) ?
                control.PropertyName :
                control.PropertyName.Substring(index + splitter.Length);
        }

        /// <summary>
        /// 指定したアイテムコードのレイヤーを検索する。
        /// </summary>
        /// <param name="composition">コンポジション。</param>
        /// <param name="itemCode">アイテムコード。</param>
        /// <returns>レイヤー。見つからなかった場合は null 。</returns>
        public static ILayer FindLayer(IComposition composition, int itemCode)
        {
            foreach (ILayer l in composition.GetLayer())
            {
                if (l.ItemCode == itemCode)
                {
                    return l;
                }
            }
            return null;
        }

        /// <summary>
        /// 矩形領域内の指定した配置位置を取得する。
        /// </summary>
        /// <param name="rectangle">矩形領域。</param>
        /// <param name="alignment">配置。</param>
        /// <returns>配置位置。</returns>
        public static PointF GetAlignmentPoint(
            RectangleF rectangle, RectangleAlignment alignment)
        {
            PointF pt = PointF.Empty;

            // X位置
            switch (alignment)
            {
                case RectangleAlignment.TopLeft:
                case RectangleAlignment.MiddleLeft:
                case RectangleAlignment.BottomLeft:
                    pt.X = rectangle.Left;
                    break;

                case RectangleAlignment.TopCenter:
                case RectangleAlignment.MiddleCenter:
                case RectangleAlignment.BottomCenter:
                    pt.X = rectangle.Left + rectangle.Width / 2;
                    break;

                case RectangleAlignment.TopRight:
                case RectangleAlignment.MiddleRight:
                case RectangleAlignment.BottomRight:
                    pt.X = rectangle.Right;
                    break;
            }

            // Y位置
            switch (alignment)
            {
                case RectangleAlignment.TopLeft:
                case RectangleAlignment.TopCenter:
                case RectangleAlignment.TopRight:
                    pt.Y = rectangle.Top;
                    break;

                case RectangleAlignment.MiddleLeft:
                case RectangleAlignment.MiddleCenter:
                case RectangleAlignment.MiddleRight:
                    pt.Y = rectangle.Top + rectangle.Height / 2;
                    break;

                case RectangleAlignment.BottomLeft:
                case RectangleAlignment.BottomCenter:
                case RectangleAlignment.BottomRight:
                    pt.Y = rectangle.Bottom;
                    break;
            }

            return pt;
        }

        /// <summary>
        /// 列挙値とその説明のペアコレクションを取得する。
        /// </summary>
        /// <typeparam name="ET">列挙型。</typeparam>
        /// <returns>列挙値とその説明のペアコレクション。</returns>
        public static IDictionary<ET, string> GetEnumDescriptions<ET>()
            where ET : struct, IConvertible
        {
            // 列挙型かチェック
            if (!typeof(ET).IsEnum)
            {
                throw new ArgumentException("ET must be enum type.");
            }

            Dictionary<ET, string> table = new Dictionary<ET, string>();

            // 列挙値ごとに処理
            FieldInfo[] infos =
                typeof(ET).GetFields(BindingFlags.Public | BindingFlags.Static);
            foreach (FieldInfo fi in infos)
            {
                // Description 属性があるなら文字列取得
                DescriptionAttribute attr =
                    Attribute.GetCustomAttribute(fi, typeof(DescriptionAttribute))
                        as DescriptionAttribute;
                string desc = (attr == null) ? null : attr.Description;

                // 追加
                table.Add((ET)Enum.Parse(typeof(ET), fi.Name), desc);
            }

            return table;
        }

        /// <summary>
        /// 列挙値とその説明の配列を取得する。
        /// </summary>
        /// <typeparam name="ET">列挙型。</typeparam>
        /// <param name="values">列挙値配列の設定先。</param>
        /// <param name="descriptions">説明文字列配列の設定先。</param>
        public static void GetEnumDescriptions<ET>(
            out ET[] values,
            out string[] descriptions)
            where ET : struct, IConvertible
        {
            IDictionary<ET, string> dict = GetEnumDescriptions<ET>();
            values = new ET[dict.Count];
            descriptions = new string[dict.Count];
            dict.Keys.CopyTo(values, 0);
            dict.Values.CopyTo(descriptions, 0);
        }

        /// <summary>
        /// 列挙の既定値を取得する。
        /// </summary>
        /// <typeparam name="ET">列挙型。</typeparam>
        /// <returns>既定の列挙値。</returns>
        public static ET GetEnumDefaultValue<ET>()
            where ET : struct, IConvertible
        {
            // 列挙型かチェック
            if (!typeof(ET).IsEnum)
            {
                throw new ArgumentException("ET must be enum type.");
            }

            // 列挙値ごとにチェック
            FieldInfo[] infos =
                typeof(ET).GetFields(BindingFlags.Public | BindingFlags.Static);
            foreach (FieldInfo fi in infos)
            {
                // DefaultValue 属性があるならその値を返す
                Attribute attr = Attribute.GetCustomAttribute(
                    fi, typeof(EnumDefaultValueAttribute));
                if (attr != null)
                {
                    return (ET)Enum.Parse(typeof(ET), fi.Name);
                }
            }

            // 見つからなかったならば通常の既定値を返す
            return new ET();
        }

        /// <summary>
        /// 列挙値配列から列挙の既定値を検索する。
        /// </summary>
        /// <typeparam name="ET">列挙型。</typeparam>
        /// <param name="values">列挙値配列。</param>
        /// <returns>既定値のインデックス。見つからなかった場合は 0 。</returns>
        public static int FindEnumDefaultValue<ET>(ET[] values)
            where ET : struct, IConvertible
        {
            ET v = GetEnumDefaultValue<ET>();
            return Array.IndexOf(values, v);
        }

        /// <summary>
        /// 列挙値配列から列挙の既定値を検索する。
        /// </summary>
        /// <typeparam name="ET">列挙型。</typeparam>
        /// <param name="values">列挙値配列。</param>
        /// <returns>既定値のインデックス。見つからなかった場合は 0 。</returns>
        public static int FindEnumDefaultValue<ET>(ET?[] values)
            where ET : struct, IConvertible
        {
            ET v = GetEnumDefaultValue<ET>();
            return Array.IndexOf(values, v);
        }

        /// <summary>
        /// 指定した時刻以前で最も近いキーフレームを取得する。
        /// </summary>
        /// <param name="keyFrames">キーフレーム配列。</param>
        /// <param name="time">時刻。</param>
        /// <returns>
        /// 指定した時刻以前で最も近いキーフレーム。
        /// キーフレームが存在しない場合は null 。
        /// </returns>
        public static KeyFrame GetPrevKeyFrame(
            KeyFrame[] keyFrames,
            double time)
        {
            KeyFrame prevKey = null;
            foreach (KeyFrame key in keyFrames)
            {
                if (key.Time <= time)
                {
                    if (prevKey == null || key.Time > prevKey.Time)
                    {
                        prevKey = key;
                    }
                }
            }
            return prevKey;
        }

        /// <summary>
        /// 指定した時刻超で最も近いキーフレームを取得する。
        /// </summary>
        /// <param name="keyFrames">キーフレーム配列。</param>
        /// <param name="time">時刻。</param>
        /// <returns>
        /// 指定した時刻超で最も近いキーフレーム。
        /// キーフレームが存在しない場合は null 。
        /// </returns>
        public static KeyFrame GetNextKeyFrame(
            KeyFrame[] keyFrames,
            double time)
        {
            KeyFrame nextKey = null;
            foreach (KeyFrame key in keyFrames)
            {
                if (key.Time > time)
                {
                    if (nextKey == null || key.Time < nextKey.Time)
                    {
                        nextKey = key;
                    }
                }
            }
            return nextKey;
        }

        /// <summary>
        /// 指定した時刻以前で最も近いキーフレームからの経過秒数を取得する。
        /// </summary>
        /// <param name="keyFrames">キーフレーム配列。</param>
        /// <param name="time">時刻。</param>
        /// <returns>
        /// 指定した時刻以前で最も近いキーフレームからの経過秒数。
        /// キーフレームが存在しない場合は time の値をそのまま返す。
        /// </returns>
        public static double CalcElapsedTime(KeyFrame[] keyFrames, double time)
        {
            // 取得時刻以前で最も近いキーフレームを探す
            KeyFrame prevKey = GetPrevKeyFrame(keyFrames, time);

            // 経過秒数を返す
            return ((prevKey == null) ? time : (time - prevKey.Time));
        }
    }
}
